#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / adapter / http / utilities
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
            'status_to_str',
            'build_host_str',
            'dummy_file_wrapper',
            'send_content_iter',
            'parse_form',
            'canonicalize_http_headers',
            'gen_http_headers',
            ]

from functools import partial

try:
    # Python 3
    from urllib.parse import parse_qs
except ImportError:
    # Python 2
    from urlparse import parse_qs

try:
    import ujson as json
except ImportError:
    import json

import six

from ...helpers.misc import smartstr, smartbytes


# This dict is pasted from Django's core/handlers/wsgi.py, with letter cases
# adjusted.
# See http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
STATUS_CODES_MAP = {
        100: 'Continue',
        101: 'Switching Protocols',
        200: 'OK',
        201: 'Created',
        202: 'Accepted',
        203: 'Non-Authoritative Information',
        204: 'No Content',
        205: 'Reset Content',
        206: 'Partial Content',
        300: 'Multiple Choices',
        301: 'Moved Permanently',
        302: 'Found',
        303: 'See Other',
        304: 'Not Modified',
        305: 'Use Proxy',
        306: 'Reserved',
        307: 'Temporary Redirect',
        400: 'Bad Request',
        401: 'Unauthorized',
        402: 'Payment Required',
        403: 'Forbidden',
        404: 'Not Found',
        405: 'Method Not Allowed',
        406: 'Not Acceptable',
        407: 'Proxy Authentication Required',
        408: 'Request Timeout',
        409: 'Conflict',
        410: 'Gone',
        411: 'Length Required',
        412: 'Precondition Failed',
        413: 'Request Entity Too Large',
        414: 'Request-URI Too Long',
        415: 'Unsupported Media Type',
        416: 'Requested Range Not Satisfiable',
        417: 'Expectation Failed',
        500: 'Internal Server Error',
        501: 'Not Implemented',
        502: 'Bad Gateway',
        503: 'Service Unavailable',
        504: 'Gateway Timeout',
        505: 'HTTP Version Not Supported',
        }

# Content types that can be parsed into a form
_FORM_CONTENT_HANDLERS = {}


def _form_content_handler(content_type):
    def _decorator_(fn):
        if content_type in _FORM_CONTENT_HANDLERS:
            raise ValueError(
                    "'%s' already registered as handler!" % (
                        content_type,
                        )
                    )

        _FORM_CONTENT_HANDLERS[content_type] = fn
        return fn
    return _decorator_


def status_to_str(status):
    '''Converts status code to its description.

    '''

    return STATUS_CODES_MAP[status]


def build_host_str(env):
    # Written according to the algorithm described in PEP 333.
    if 'HTTP_HOST' in env:
        return smartstr(env['HTTP_HOST'], 'utf-8', 'replace')
    else:
        server_name_str = smartstr(env['SERVER_NAME'], 'utf-8', 'replace')
        server_port = int(env['SERVER_PORT'])
        host_str_list = [server_name_str, ]

        if env['wsgi.url_scheme'] == 'https':
            if server_port != 443:
                host_str_list.append(six.text_type(server_port))
        else:
            if server_port != 80:
                host_str_list.append(six.text_type(server_port))

        return ':'.join(host_str_list)


def dummy_file_wrapper(fp, blk_sz=None):
    blk_sz = blk_sz if blk_sz is not None else 4096
    do_read = partial(fp.read, blk_sz)
    chunk = do_read()
    while len(chunk) > 0:
        yield chunk
        chunk = do_read()
    fp.close()


def send_content_iter(content, enc):
    if isinstance(content, six.string_types):
        yield smartbytes(content, enc, 'replace')
    elif isinstance(content, six.binary_type):
        yield content
    else:
        for chunk in content:
            # encode and send the chunk using response.encoding
            yield smartbytes(chunk, enc, 'replace')


def parse_form(ctype, content):
    if ctype in _FORM_CONTENT_HANDLERS:
        return _FORM_CONTENT_HANDLERS[ctype](content)

    return None


@_form_content_handler('application/x-www-form-urlencoded')
def _parse_urlencoded_form(content):
    form = parse_qs(content)

    # eliminate all those 1-element lists
    for k in six.iterkeys(form):
        if len(form[k]) == 1:
            form[k] = form[k][0]

    return form


@_form_content_handler('application/json')
def _parse_json_form(content):
    return json.loads(content)


def canonicalize_http_headers(header_obj):
    if isinstance(header_obj, list):
        return header_obj
    elif isinstance(header_obj, dict):
        # most probably dict or OrderedDict, make it a plain list
        return list(six.iteritems(header_obj))
    elif isinstance(header_obj, tuple):
        return list(header_obj)

    raise ValueError(
            'unrecognized HTTP headers object: %s' % (
                repr(header_obj),
                ))


def gen_http_headers(response, __Content_Length=str('Content-Length')):
    status_code, enc = response.status, response.encoding

    status_line = str('%d %s' % (
            status_code,
            status_to_str(status_code),
            ))

    # ensure all header contents are bytes
    headers = []

    # insert a Content-Length along if response is not raw file and contains
    # a body
    if not response.is_raw_file and not response._dont_render:
        headers.append((
                __Content_Length,
                str(len(response.content)),
                ))

    for k, v in response.http_headers:
        k_bytes, v_bytes = smartbytes(k, enc), smartbytes(v, enc)
        k_str, v_str = str(k_bytes), str(v_bytes)
        headers.append((k_str, v_str, ))

    return status_line, headers


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / central registry / config loader
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Configuration loader
~~~~~~~~~~~~~~~~~~~~

As much of weiyu's functionality relies on the registries, there should be
a convenient way of pre-loading configuration data before performing any
significant actions. This module serves just that purpose, providing several
configuration formats free to choose from; of course, you are also free to
roll your own loader by inheriting :class:`.BaseConfig` and implement the
``_do_loads`` method. At present, only JSON and Python pickles are
available for use; more is expected to be added.


Basics
------

Global configuration files which is used by
:meth:`BaseConfig.populate_central_regs` to populate (possibly) multiple
registries must have a Pythonized format like this::

    {
        u'registry.name.1': {
            u'class': u'ClassNameOfRegistry', # must be bytes or unicode
            u'no_update': True or False,
            u'data': {
                # this is the actual configuration data
                u'spam': u'foo',
                u'eggs': 42,
                u'etc': u'etc',
                }
            },

        # optionally more registries, or directives to follow
    }

The ``class`` attribute must specify one of the registry types "exported"
in :mod:`weiyu.registry.classes`; this limitation may be lifted in future
versions of weiyu.

The ``no_update`` boolean, whose value is directly passed as the
``nodup`` parameter to :func:`weiyu.registry.provider.request`,
indicates that an exception should be raised when a registry with the
same name already exists (possibly initialized to an empty one by its
associated module). It really should be set to ``False`` for populating
registries initialized by modules. Only set this to ``True`` when you are
sure that no module will try to create the registry before you load the
configuration.


Preprocessing directives
------------------------

Directives are just sections in the config file with names starting with
``$$``. They are then parsed by :meth:`BaseConfig.process_directives`.

Currently there is only one directive implemented: the (top-level) include
directive ``$$include``. Its parameter is a ``list`` of files to source:
content of the files are read and processed in order, overwriting values
if they already exist. It does not support including files into a variable,
for example, at the moment. Also it does not detect circular includes, any
such situation the loader will loop infinitely and eventually crash with a
:exc:`MemoryError`. Don't try that.


Classes
-------

'''

from __future__ import unicode_literals, division

__all__ = [
        'BaseConfig',
        'JSONConfig',
        'PickleConfig',
        ]

from os.path import abspath, dirname, splitext
from os.path import join as pathjoin
from functools import wraps
import abc
import json

import six
from six.moves import cPickle as pickle

from .classes import VALID_REGISTRY_TYPES
from .provider import request

# $$directives
DIRECTIVE_CLASS_OVERRIDE = '$$class'
DIRECTIVE_UPDATE_OVERRIDE = '$$nodup'
DIRECTIVE_INCLUDE = '$$include'

# config file extension handler registry
_EXTENSION_HANDLERS = {}


def handler_for_ext(ext):
    def _decorator_(thing):
        _EXTENSION_HANDLERS[ext] = thing
        return thing
    return _decorator_


class BaseConfig(six.with_metaclass(abc.ABCMeta)):
    '''Base class for prepopulated config storages. It has an interface
    inspired by ``pickle``: methods such as :meth:`loads` and :meth:`load`
    exist and more or less behaves like the standard library counterpart.

    The :meth:`_do_loads` method is abstract, it is up to you to decide the
    way configuration files map to Python object in a subclass.

    '''

    @staticmethod
    def get_config(path):
        ext = splitext(path)[1]

        try:
            cls = _EXTENSION_HANDLERS[ext]
        except KeyError:
            raise ValueError(
                    "cannot handle config file '%s' with extension '%s'" % (
                        path,
                        ext,
                        ),
                    )

        return cls(path)

    def __init__(self, path=None):
        '''Constructor function.

        :param path: Path to config file. Defaults to ``None``; If this is
        the case, data may be loaded by directly calling :meth:`loads` or
        :meth:`load` with the appropriate type of data source.

        '''

        self.path, self.data = path, None

    @abc.abstractmethod
    def _do_loads(self, s, *args, **kwargs):
        '''Constructs a Python object from input. This method is abstract.

        :param s: The input string.

        .. note::

            Don't do preprocessing in this method; they are to be done by
            :meth:`loads` and the other interface methods. Just construct
            the "raw" deserialized form.

        '''

        pass

    def loads(self, s, *args, **kwargs):
        '''Loads configuration from a string.
        '''

        raw_data = self._do_loads(s, *args, **kwargs)
        self.data = self.process_directives(raw_data)
        return self.data

    def load(self, fp=None, *args, **kwargs):
        '''Loads configuration from a file-like object with a ``read`` method.

        .. note::

            The implementation at this point is naive. It just tries to read
            the whole input, then feeds the string into :meth:`loads`. Beware
            if the input is potentially very large.

        '''

        # XXX this function's a naive wrapper to get rid of some duplication

        if fp is None:
            # invoke ``self.load_from_path`` to read the file specified
            # by ``self.path``
            return self.load_from_path(None)

        # Unicode cleanness
        return self.loads(
                fp.read().decode('utf-8', 'replace'),
                *args, **kwargs
                )

    def load_from_path(self, path=None):
        '''Loads configuration from path ``path``; if ``path`` is ``None``,
        try to use ``self.path``. If that value is also ``None``, a
        :exc:`ValueError` is ``raise``\ d.

        '''

        if path is None:
            if self.path is None:
                raise ValueError('no usable path configured')
            path = self.path

        return self._do_load_from_path(path)

    def _do_load_from_path(self, path):
        with open(path, 'rb') as fp:
            return self.load(fp)

    def _do_inject_include_files(self, includes):
        tmp = {}

        # All include paths should be relative to the config file currently
        # under preprocessing.
        this_path = dirname(abspath(self.path))

        # print("$$include: base path for '%s' is '%s'" % (
        #         self.path,
        #         this_path,
        #         ))

        for path in includes:
            # XXX FIXME: Infinite includes is possible!!
            # NOTE Security is important here, so paths should be
            # at least canonicalized.

            real_path = abspath(pathjoin(this_path, path))
            # print("$$include: '%s' resolved to '%s'" % (path, real_path, ))

            tmp.update(BaseConfig.get_config(real_path).load())

        return tmp

    def _do_includes(self, dct):
        # top-level includes
        if DIRECTIVE_INCLUDE in dct:
            include_file_or_list = dct.pop(DIRECTIVE_INCLUDE)

            if isinstance(include_file_or_list, six.text_type):
                include_list = [include_file_or_list, ]
            elif isinstance(include_file_or_list, six.binary_type):
                include_list = [include_file_or_list, ]
            elif isinstance(include_file_or_list, (list, tuple, )):
                include_list = include_file_or_list
            else:
                raise ValueError(
                        'unknown object for $$include directive: %s' % (
                            repr(include_file_or_list),
                            ),
                        )

            dct.update(self._do_inject_include_files(include_list))

        # includes inside sub-keys
        for k in six.iterkeys(dct):
            if isinstance(dct[k], dict):
                # recurse into ANY sub-dictionary
                dct[k] = self._do_includes(dct[k])

        return dct

    def process_directives(self, data):
        # process all $$include's
        # print("Preprocessing config file '%s'" % (self.path, ))
        data = self._do_includes(data)

        return data

    def ensure_data_presence(self):
        '''When called, checks if ``self.data`` is ``None``. If it is the
        case, calls :meth:`load_from_path` to load configuration from
        ``self.path``. Note that this method does NOT ensure the
        *consistency* between the stored state and the config file's data;
        it just checks for presence.

        '''

        if self.data is None:
            self.load_from_path(None)

    def populate_registry(self, registry):
        '''Populates a single registry instance. For filling the global
        registry, use :meth:`populate_central_regs`.

        '''

        self.ensure_data_presence()
        for k, v in six.iteritems(self.data):
            registry.register(k, v)

    def populate_central_regs(self):
        '''Fills the central registry, i.e. the
        :class:`.classes.RegistryRegistry` singleton. Config data format
        used for this method is described above.

        '''

        self.ensure_data_presence()
        for k, v in six.iteritems(self.data):
            # k is registry name, v is a dict
            # v's format is as follows:
            # {['$$class': '<class name of registry (optional)>',]
            #  ['$$nodup': True/False (optional),]
            #  key1: val1,
            #  key2: val2,
            #  etc: etc,
            #  }
            # first a little type sanity check
            if not isinstance(k, six.string_types):
                raise ValueError(
                        'registry name must be string; got %s (of type %s)'
                        % (repr(k), repr(type(k)), )
                        )

            if not isinstance(v, dict):
                raise ValueError(
                        'pythonized config entry must be dict; got type %s'
                        % (repr(type(v)), )
                        )

            # extract class and updating behavior
            if DIRECTIVE_CLASS_OVERRIDE in v:
                class_name = v.pop(DIRECTIVE_CLASS_OVERRIDE)
            else:
                # hard-code to be a string-keyed registry
                class_name = 'UnicodeRegistry'

            if DIRECTIVE_UPDATE_OVERRIDE in v:
                nodup = v.pop(DIRECTIVE_UPDATE_OVERRIDE)
            else:
                # allow updating existing registries through conf by default
                nodup = False

            # override field type check
            if not isinstance(class_name, six.string_types):
                raise ValueError('registry class name not of string type')

            if not isinstance(nodup, bool):
                raise ValueError('no_update property must be boolean')

            # class validation
            try:
                cls = VALID_REGISTRY_TYPES[class_name]
            except KeyError:
                raise ValueError("registry class name '%s' not valid"
                        % (class_name, )
                        )

            # format check (finally...) passed, go on with actually updating
            # NOTE: obviously registries cannot have duplicate names,
            # here we'll rely on RegistryBase's behavior to ensure this :P
            reg = request(k, autocreate=True, nodup=nodup, klass=cls)

            # populate the dict
            # NOTE: We don't use update() here, because filling values one by
            # one would trigger validation and normalization for each of the
            # k-v pairs, and sanity is what we need.
            for reg_k, reg_v in six.iteritems(v):
                reg[reg_k] = reg_v


@handler_for_ext('.json')
class JSONConfig(BaseConfig):
    '''JSON config backend.'''

    def _do_loads(self, s, *args, **kwargs):
        dump = json.loads(s, *args, **kwargs)
        # TODO: some canonicalization or such
        return dump


@handler_for_ext('.pickle')
class PickleConfig(BaseConfig):
    '''Python pickle config backend.'''

    def _do_loads(self, s, *args, **kwargs):
        return pickle.loads(s, *args, **kwargs)


@handler_for_ext('.yml')
@handler_for_ext('.yaml')
class YAMLConfig(BaseConfig):
    '''YAML config backend.'''

    def _do_loads(self, s, *args, **kwargs):
        import yaml
        from yaml import Loader, SafeLoader

        # Force Unicode string output according to this SO question
        # 2890146/how-to-force-pyyaml-to-load-strings-as-unicode-objects
        def construct_yaml_str(self, node):
            # Override the default string handling function
            # to always return unicode objects
            return self.construct_scalar(node)

        _STR_TAG = 'tag:yaml.org,2002:str'
        Loader.add_constructor(_STR_TAG, construct_yaml_str)
        SafeLoader.add_constructor(_STR_TAG, construct_yaml_str)

        # TODO: optionally utilize C acceleration if available
        return yaml.load(s, *args, **kwargs)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / router / reverser
#
# Copyright (C) 2013-2014 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'reverser_for_router',
        'Reverser',
        ]

import six

_REVERSER_CACHE = {}


def reverser_for_router(router):
    key = router.name
    if key is None:
        raise ValueError('only named routers can have reverser at present')

    try:
        return _REVERSER_CACHE[key]
    except KeyError:
        reverser = Reverser(router)
        _REVERSER_CACHE[key] = reverser
        return reverser


class Reverser(object):
    def __init__(self, router):
        self.router = router
        self.map = router.reverse_map
        self.pat_cache = {}

    def _resolve_scope(self, endpoint, __cache={}):
        # split the possibly scoped endpoint name into components
        try:
            # scope, name
            return __cache[endpoint]
        except KeyError:
            pass

        try:
            colon = endpoint.index(':')
            scope, name = endpoint[:colon], endpoint[colon + 1:]
        except ValueError:
            scope, name = '', endpoint

        # memoize
        result = __cache[endpoint] = (scope, name, )
        return result

    def signature(self, endpoint):
        # look up the endpoint, memoized.
        try:
            # pat_str, pat_vars
            return self.pat_cache[endpoint]
        except KeyError:
            pass

        scope, name = self._resolve_scope(endpoint)

        # look up the scope first
        try:
            scope_map = self.map[scope]
        except KeyError:
            # TODO: a NoReverseMatch here, as Django did?
            raise ValueError(
                    "No scope named '%s' in router type '%s'" % (
                        scope,
                        typ,
                        ))

        # get out the pattern signature
        try:
            pat_str, pat_vars = scope_map[name]
        except KeyError:
            raise ValueError(
                    "No endpoint '%s' exposed in scope '%s' of "
                    "router type '%s'" % (
                        name,
                        scope,
                        typ,
                        ))

        # memoize the result
        result = self.pat_cache[endpoint] = (pat_str, pat_vars, )
        return result

    def reverse(self, endpoint, **kwargs):
        pat_str, pat_vars = self.signature(endpoint)

        # verify the parameters
        given_vars = set(six.iterkeys(kwargs))
        if given_vars != pat_vars:
            # parameter mismatch
            raise ValueError('Parameter mismatch')

        # successful match, construct the result
        return pat_str % kwargs


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / router / package
#
# Copyright (C) 2012-2014 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'router_hub',
        ]

import os
import errno

import six

from ..helpers.hub import BaseHub
from ..helpers.modprober import ModProber
from ..registry.classes import UnicodeRegistry

# this does not cause circular import
from .config.parser import parse_config

from .reverser import reverser_for_router

PROBER = ModProber('weiyu.router', '%srouter')


class RouterHub(BaseHub):
    registry_name = 'weiyu.router'
    registry_class = UnicodeRegistry
    handlers_key = 'handlers'

    def __init__(self):
        super(RouterHub, self).__init__()

        if 'endpoints' not in self._reg:
            self._reg['endpoints'] = {}

        if 'routers' not in self._reg:
            self._reg['routers'] = {}

        if 'classes' not in self._reg:
            self._reg['classes'] = {}

        # cache the references
        self._routers = self._reg['routers']
        self._endpoints = self._reg['endpoints']
        self._classes = self._reg['classes']

    def endpoint(self, typ, name):
        '''decorator for registering routing end points.'''

        def _decorator_(fn):
            if typ not in self._endpoints:
                # this type not already registered, probably not inited yet
                # let's give it a sensible default
                self._endpoints[typ] = {}
            self._endpoints[typ][name] = fn
            return fn
        return _decorator_

    def register_router_class(self, name):
        '''Decorator to make a class available for routing.

        This decorator is mainly intended for internal use, to give short
        names to the router classes.

        '''

        def _decorator_(cls):
            if name in self._classes:
                raise ValueError(
                        'duplicate router register name: \'%s\'' % (name, )
                        )

            self._classes[name] = cls
            return cls
        return _decorator_

    def register_router(self, router):
        # keep a reference to the router
        typ = router.name
        if typ is None:
            raise ValueError(
                    'only named routers can be registered this way'
                    )

        self._routers[typ] = router

        # also reserve a slot in endpoints dict, if one is not already
        # set up
        if typ not in self._endpoints:
            self._endpoints[typ] = {}

        # register the router's dispatch method as handler, and we're done
        # here we'd use dry dispatch instead...
        # But first construct a shim removing the hub parameter...
        @self.register_handler(typ)
        def _routing_shim_(hub, *args, **kwargs):
            return router.dry_dispatch(*args, **kwargs)

    def dry_dispatch(self, typ, querystr, *args):
        # typically used with args=(request, ) inside the framework
        # TODO: is it really useful to allow passing kwargs also?
        return self.do_handling(typ, querystr, *args)

    def reverser_for(self, router_name):
        return reverser_for_router(self._routers[router_name])

    def _probe_router_class(self, cls_name):
        try:
            return self._classes[cls_name]
        except KeyError:
            pass

        try:
            PROBER.modprobe(cls_name)
        except ImportError:
            raise RuntimeError(
                    'request of unknown router class \'%s\'' % (
                        cls_name,
                        ),
                    )

        # Assume that module has actually registered the wanted class...
        assert cls_name in self._classes
        return self._classes[cls_name]

    def _do_init_router(
            self,
            typ,
            routing_rules,
            lvl,
            parent_info,
            _list_types=(list, tuple, ),
            ):
        # recursive algorithm, watch out d-:
        # typ is not really useful except checking against endpoint reg
        #
        # let's construct the desired target initializer out of the
        # pattern-to-(endpoint-or-router) list
        #
        # Attribute processing.
        attrib_list = routing_rules[0]
        this_file, inherited_klass, inherited_renderer, scope, host = (
                parent_info['__file__'],
                parent_info['inherited_klass'],
                parent_info['inherited_renderer'],
                parent_info['scope'],
                parent_info['host'],
                )
        include_path, cls_name = None, None

        if not isinstance(attrib_list, _list_types):
            # Always use a list for attributes.
            attrib_list = [attrib_list]

        for attrib in attrib_list:
            # Router class spec.
            if '=' not in attrib:
                # The class spec attrib does not contain '='.
                # This flaky way of determining class spec can be used
                # at least for now, as no other parameter-less attributes
                # exist so far.
                cls_name = attrib
                continue

            # Router properties.
            k, v = attrib.split('=', 1)

            if k == 'renderer':
                # case: renderer=xxx
                # record the renderer to inherit
                inherited_renderer = v
            elif k == 'include':
                # case: include=xxx
                # it must appear as the only attribute, if used!
                if len(attrib_list) > 1:
                    # throw an exception.
                    raise RuntimeError(
                            'include directive must appear on its own'
                            )
                include_path = v
            elif k == 'scope':
                # case: scope=xxx
                scope = v
            elif k == 'default-type':
                # case: default-type=xxx
                inherited_klass = v
            elif k == 'host':
                # case: host=xxx
                host = v
            else:
                # Unknown attribute, ignore it for compatibility with future
                # versions.
                # TODO: log warnings
                continue

        # Process include.
        if include_path is not None:
            # There must be no other routing rules described... or we would
            # not know what to include.
            if len(routing_rules) > 1:
                raise RuntimeError(
                        'included router description should not contain '
                        'any rules'
                        )

            # Resolve include path.
            if this_file is not None:
                # Resolve relative to the current file.
                this_dir = os.path.dirname(this_file)
            else:
                # Current file path is unavailable, resolve relative to the
                # current working directory (old behavior).
                this_dir = os.getcwdu()

            include_path_resolved = os.path.abspath(
                    os.path.join(
                        this_dir,
                        include_path,
                        ))

            # Allow for extension (.URLfile) omission.
            #
            # Since it is not the default any more, .txt is not considered
            # here. The worst thing can happen is a failed open() call, so
            # no worries.
            #
            # I decided to not allow funky things like including something
            # like 'foo.URLfile.URLfile'...
            if not include_path_resolved.endswith('.URLfile'):
                # Likely the filename extension is omitted, add it back.
                suffixed_path = include_path_resolved + '.URLfile'

                # Try the suffixed version first...
                try:
                    return self.init_router_from_config(typ, suffixed_path)
                except IOError as e:
                    # Only ignore ENOENT here.
                    if e.errno != errno.ENOENT:
                        raise

            return self.init_router_from_config(typ, include_path_resolved)

        # Load the requested (built-in) router class.
        #
        # Custom router classes should be registered before config load, like
        # before.
        #
        # TODO: allow registration of custom modules in modprober mechanism

        # Override with inherited class name if not specified already.
        # This works even inside the same router node.
        cls_name = cls_name if cls_name is not None else inherited_klass
        cls = self._probe_router_class(cls_name)

        # Process rules.
        result_rules = []
        for pattern, target_spec, extra_data in routing_rules[1:]:
            if isinstance(target_spec, _list_types):
                # this is a router... recursively construct a router out
                # of it
                my_info = {
                        '__file__': this_file,
                        'inherited_klass': inherited_klass,
                        'inherited_renderer': inherited_renderer,
                        'scope': scope,
                        'host': host,
                        }
                tgt = self._do_init_router(typ, target_spec, lvl + 1, my_info)
            elif isinstance(target_spec, six.string_types):
                # target is endpoint... check against endpoint registry
                # this is where typ is used
                # enforce a little bit of encoding requirement, to make
                # messed up encoding problem surface fast
                tgt = self._endpoints[typ][six.text_type(target_spec)]
            else:
                # unrecognized target specification, pass it thru as is
                tgt = target_spec

            # process special data
            if extra_data is not None:
                # render_in
                if extra_data.get('render_in', None) == 'inherit':
                    # inherit renderer from parent
                    extra_data['render_in'] = inherited_renderer
            else:
                extra_data = {}

            # remember the target spec for the URL reverser
            extra_data['target_spec'] = target_spec

            # add a rule
            result_rules.append((pattern, tgt, extra_data, ))

        # construct a XxxRouter object with the routing rules just created
        # if toplevel router, assign it a name equal to typ
        return cls(
                result_rules,
                name=typ if lvl == 0 else None,
                scope=scope,
                host=host,
                )

    def init_router(self, typ, routing_rules, filename=None):
        return self._do_init_router(
                typ,
                routing_rules,
                0,
                {
                    '__file__': filename,
                    'inherited_klass': None,
                    'inherited_renderer': None,
                    'scope': '',
                    'host': None,
                    },
                )

    def init_router_from_config(self, typ, filename):
        config = parse_config(filename)
        return self.init_router(typ, config, filename)


router_hub = RouterHub()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / framework-level initialization / package
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'load_router',
        'load_config',
        'load_views',
        'boot',
        ]

import os
import inspect
import warnings

import six

from ..adapters import adapter_hub
from ..db import db_hub
from ..router import router_hub
from ..registry.classes import UnicodeRegistry
from ..registry.loader import BaseConfig
from ..registry.provider import request as regrequest
from .viewloader import ViewLoader

# XXX Force load of HTTPSessionMiddleware
from ..adapters.http import base
del base

CONFIG_LOADED = False


def _ensure_site_registry():
    return regrequest(
            'site',
            autocreate=True,
            nodup=False,
            klass=UnicodeRegistry,
            )


def _do_load_router(typ, filename):
    router = router_hub.init_router_from_config(typ, filename)
    return router_hub.register_router(router)


def load_router(typ, filename=None):
    if filename is not None:
        # configured by app stub
        return _do_load_router(typ, filename)

    reg_site = _ensure_site_registry()
    if 'urls' in reg_site:
        # configured by registry
        return _do_load_router(typ, reg_site['urls'])

    # probe for different defaults
    if os.path.exists('Rain.d/root.URLfile'):
        if os.path.exists('urls.txt'):
            warnings.warn(
                    'Rain.d/root.URLfile overrides urls.txt, please migrate '
                    'and delete the old URL routing files to avoid any '
                    'possible inconsistencies',
                    DeprecationWarning,
                    )
        return _do_load_router(typ, 'Rain.d/root.URLfile')

    if os.path.exists('urls.txt'):
        warnings.warn(
                'The filename \'urls.txt\' for URL routing is '
                'deprecated, please move to Rain.d/root.URLfile (no other '
                'changes needed)',
                DeprecationWarning,
                )
        return _do_load_router(typ, 'urls.txt')

    raise RuntimeError('no default URL routing configuration found')


def _do_load_config(path):
    global CONFIG_LOADED
    if CONFIG_LOADED:
        return

    # this has no return value as of now, safe to discard returned None
    BaseConfig.get_config(path).populate_central_regs()

    # Refresh the hubs' internal cached references, this MUST be done
    # after config has loaded.
    db_hub._init_refresh_map()

    CONFIG_LOADED = True


def load_config(path=None):
    if path is not None:
        _do_load_config(path)
    else:
        # Default config file.
        if os.path.exists('Rain.d/config.yml'):
            # 1. Rain.d ((again) new default)
            if any(
                    os.path.exists(f)
                    for f in ('Rainfile.yml', 'conf.yml', 'conf.json')
                    ):
                warnings.warn(
                        'Rain.d/config.yml overrides the config files with '
                        'old default names, please migrate and delete the '
                        'old configuration files to avoid any possible '
                        'inconsistencies',
                        DeprecationWarning,
                        )
            _do_load_config('Rain.d/config.yml')
        elif os.path.exists('Rainfile.yml'):
            # 2. Rainfile (old default)
            warnings.warn(
                    'The filename \'Rainfile.yml\' for configuration is '
                    'deprecated, please move to Rain.d/config.yml (no other '
                    'changes needed)',
                    DeprecationWarning,
                    )
            _do_load_config('Rainfile.yml')
        elif os.path.exists('conf.yml'):
            # 3. conf.yml (old^2 default)
            warnings.warn(
                    'The filename \'conf.yml\' for configuration is '
                    'deprecated, please move to Rain.d/config.yml (no other '
                    'changes needed)',
                    DeprecationWarning,
                    )
            _do_load_config('conf.yml')
        elif os.path.exists('conf.json'):
            # 4. conf.json (VERY VERY old default, since weiyu's inception)
            warnings.warn(
                    'JSON configuration stored in \'conf.json\' is LONG '
                    'deprecated. Please consider migrating your '
                    'configurations to YAML for better readability and ease '
                    'of maintenance, then store it in Rain.d/config.yml',
                    DeprecationWarning,
                    )
            _do_load_config('conf.json')
        else:
            raise RuntimeError('no default configuration found')


def load_views(path_or_config):
    if isinstance(path_or_config, six.string_types):
        # this is a path to views.json
        return ViewLoader().fileconfig(path_or_config)()

    # direct config by dict
    return ViewLoader(path_or_config)()


def boot(
        conf_path=None,
        views_path=None,
        router_type='http',
        root_router_file=None,
        ):
    # initialize registries, views, router, in that order
    # registry
    load_config(conf_path)

    # make an empty site registry if not present
    reg_site = _ensure_site_registry()

    # determine if ViewLoader is configured by registry
    if views_path is not None:
        # external file configured by app stub
        load_views(views_path)
    else:
        # extract config from site registry
        if 'views' in reg_site:
            view_config = reg_site['views']
            load_views(view_config)
        else:
            # old behavior, which is an implicit (specified via default)
            # views.json, kept for compatibility
            warnings.warn(
                    'Implicit views.json is discouraged, please put the '
                    'view config directly into main config file',
                    PendingDeprecationWarning,
                    )
            load_views('views.json')

    # init router
    load_router(router_type, root_router_file)

    # register middlewares according to config
    middleware_decl = (
            reg_site['middlewares']
            if 'middlewares' in reg_site
            else {}
            )

    def call_register_middleware(kind):
        middlewares = middleware_decl.get(kind, [])
        adapter_hub.register_middleware_chain(
                router_type,
                kind,
                middlewares,
                )

    call_register_middleware('pre')
    call_register_middleware('post')


def inject_app(app_type='wsgi', var='application', *args, **kwargs):
    # Make app
    boot(*args, **kwargs)
    app = adapter_hub.make_app(app_type)

    # Then inject the object into the caller's global namespace
    parentframe = inspect.getouterframes(inspect.currentframe())[1][0]
    parentframe.f_globals[var] = app


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:
